////////////////////////////////////////////////////////////////////////////////
/// GENERIC METHODS
////////////////////////////////////////////////////////////////////////////////


// Convert given string to Title Case
String.prototype.toTitleCase = function() {
    return this.replace(/\w\S*/g, function(txt) { return txt.charAt(0).toUpperCase() + txt.substr(1).toLowerCase(); });
};




////////////////////////////////////////////////////////////////////////////////
/// EVERYTHING BELOW SHOULD BE WITHIN WINDOW.CHARTS NAMESPACE
////////////////////////////////////////////////////////////////////////////////

// Global Variables
window.charts = {};
window.charts.chart_list = [];
window.charts.debug = false;
window.charts.default_colors = ['#000000', 'red', 'blue', '#1c61ae', '#3a9dc8', '#af792b', '#cfac2e'];
window.charts.default_ma_colors = ['#000000', 'red', 'blue'];
window.charts.default_overlay_colors = ['#008A06'];


////////////////////////////////////////////////////////////////////////////////
/// ANNOTATION SYMBOLS
////////////////////////////////////////////////////////////////////////////////

// Annotator function for surprise
window.charts.drawSurpriseAnnotations = function(value) {
    if (value > 50) {
        // up arrow
        return { path: "M5 1 L9 6 L 1 6 L5 1 M4 6 L4 30 L6 30 L6 6", stroke: "transparent", strokewidth: "0px", fill: "green", height: -30, width: 10 };
    } else if (value >= 40) {
        // up arrow
        return { path: "M5 1 L9 6 L 1 6 L5 1 M4 6 L4 10 L6 10 L6 6 M4 11 L4 15 L6 15 L6 11 L4 11 M4 16 L4 20 L6 20 L6 16 L4 16 M4 21 L4 25 L6 25 L6 21 L4 21 M4 26 L4 30 L6 30 L6 26 L4 26", stroke: "transparent", strokewidth: "0px", fill: "green", height: -30, width: 10 };
    } else if (value >= 30) {
        // up arrow
        return { path: "M5 1 L9 6 L 1 6 L5 1 M4 6 L4 10 L6 10 L6 6 M4 11 L4 15 L6 15 L6 11 L4 11 M4 16 L4 20 L6 20 L6 16 L4 16 M4 21 L4 25 L6 25 L6 21 L4 21", stroke: "transparent", strokewidth: "0px", fill: "green", height: -25, width: 10 };
    } else if (value >= 20) {
        // up arrow
        return { path: "M5 1 L9 6 L 1 6 L5 1 M4 6 L4 10 L6 10 L6 6 M4 11 L4 15 L6 15 L6 11 L4 11 M4 16 L4 20 L6 20 L6 16 L4 16", stroke: "transparent", strokewidth: "0px", fill: "green", height: -20, width: 10 };
    } else if (value >= 10) {
        // up arrow
        //return {path: "M5 1 L9 6 L 1 6 L5 1 M4 6 L4 10 L6 10 L6 6 M4 11 L4 15 L6 15 L6 11 L4 11", stroke: "transparent", strokewidth: "0px", fill: "green", height: -15, width: 10};
        return { path: "M5 1 L9 6 L 1 6 L5 1 M4 6 L4 10 L6 10 L6 6", stroke: "transparent", strokewidth: "0px", fill: "green", height: -13, width: 10 };
    } else if (value >= 0) {
        // up arrow
        return { path: "M5 1 L9 6 L 1 6 L5 1 M4 6 L4 10 L6 10 L6 6", stroke: "transparent", strokewidth: "0px", fill: "green", height: -10, width: 10 };
    } else if (value >= -10) {
        // down arrow
        return { path: "M5 10 L9 4 L1 4 M4 4 L4 1 L6 1 L6 4", stroke: "transparent", strokewidth: "0px", fill: "red", height: 10, width: 10 };
    } else if (value >= -20) {
        // down arrow
        //return {path: "M5 15 L9 9 L1 9 M4 9 L4 6 L6 6 L6 9 M4 4 L4 1 L6 1 L6 4", stroke: "transparent", strokewidth: "0px", fill: "red", height: 15, width: 10};
        return { path: "M5 15 L9 9 L1 9 M4 9 L4 6 L6 6 L6 9", stroke: "transparent", strokewidth: "0px", fill: "red", height: 18, width: 10 };
    } else if (value >= -30) {
        // down arrow
        return { path: "M5 20 L9 14 L1 14 M4 14 L4 11 L6 11 L6 14 M4 9 L4 6 L6 6 L6 9 M4 4 L4 1 L6 1 L6 4", stroke: "transparent", strokewidth: "0px", fill: "red", height: 20, width: 10 };
    } else if (value >= -40) {
        // down arrow
        return { path: "M5 25 L9 19 L1 19 M4 19 L4 16 L6 16 L6 19 M4 14 L4 11 L6 11 L6 14 M4 9 L4 6 L6 6 L6 9 M4 4 L4 1 L6 1 L6 4", stroke: "transparent", strokewidth: "0px", fill: "red", height: 25, width: 10 };
    } else if (value >= -50) {
        // down arrow
        return { path: "M5 30 L9 24 L1 24 M4 24 L4 21 L6 21 L6 24 M4 19 L4 16 L6 16 L6 19 M4 14 L4 11 L6 11 L6 14 M4 9 L4 6 L6 6 L6 9 M4 4 L4 1 L6 1 L6 4", stroke: "transparent", strokewidth: "0px", fill: "red", height: 30, width: 10 };
    } else {
        // down arrow
        return { path: "M5 30 L9 24 L1 24 M4 24 L4 1 L6 1 L6 24", stroke: "transparent", strokewidth: "0px", fill: "red", height: 30, width: 10 };
    }
};

// Annotator function for transparent circles
window.charts.drawCircle = function(value) {
    return { path: "M5 1 L8 3 L9 5 L8 7 L5 9 L2 7 L1 5 L2 3", stroke: "transparent", strokewidth: "6px", fill: "transparent", height: 5, width: 11 };
};

// Annotator function for fundamenta l charts
window.charts.drawAnnotaionFundamental1 = function(value) {
    //return {path: "M3 1 L5 3 L3 5 L1 3", stroke: "transparent", strokewidth: "6px", fill: "#000", height: 3, width: 5};
    return { path: "M5 1 L8 3 L9 5 L8 7 L5 9 L2 7 L1 5 L2 3", stroke: "transparent", strokewidth: "6px", fill: "transparent", height: 5, width: 11 };
};

// Annotator function for fundamenta 2 charts
window.charts.drawAnnotaionFundamental2 = function(value) {
    //return {path: "M3 1 L5 3 L3 5 L1 3", stroke: "transparent", strokewidth: "6px", fill: "#f0f", height: 3, width: 5};
    return { path: "M5 1 L8 3 L9 5 L8 7 L5 9 L2 7 L1 5 L2 3", stroke: "transparent", strokewidth: "6px", fill: "transparent", height: 5, width: 11 };
}

// Annotator function for broker recommendation 'X's
window.charts.drawX = function(value) {
    return { path: "M0 0 L6 6 M0 6 L6 0", stroke: "green", strokewidth: "2px", fill: "transparent", height: 3, width: 3 };
};

// Hexagonal endpoint style
window.charts.hexagonalEndpoints = {
    "left": function(textWidth) {
        return {
            path: "-10,7 -3,0 " + (textWidth + 8) + ",0 " + (textWidth + 15) + ",7 " + (textWidth + 8) + ",15 -3,15 -10,7",
            width: textWidth + 30,
            height: 15,
            textOffsetLeft: 0,
            textOffsetTop: 0
        };
    },
    "right": function(textWidth) {
        return {
            path: "-10,7 -3,0 " + (textWidth + 8) + ",0 " + (textWidth + 15) + ",7 " + (textWidth + 8) + ",15 -3,15 -10,7",
            width: textWidth + 30,
            height: 15,
            textOffsetLeft: 0,
            textOffsetTop: 0
        };
    }
};

// Small arrow endpoint style
window.charts.smallArrowEndpoints = {
    "left": function(textWidth) {
        return {
            //path: "0,0 8,0 15,7 8,15 0,15 0,0",
            path: "0,0 4,0 8,4 4,8 0,8 0,0",
            width: 8,
            height: 8,
            textOffsetLeft: 0,
            textOffsetTop: 0
        };
    },
    "right": function(textWidth) {
        return {
            //path: "0,7 7,0 15,0 15,15 7,15, 0,7",
            path: "0,4 4,0 8,0 8,8 4,8, 0,4",
            width: 8,
            height: 8,
            textOffsetLeft: 0,
            textOffsetTop: 0
        };
    }
};



////////////////////////////////////////////////////////////////////////////////
/// ANNOTATION TOOLTIPS
////////////////////////////////////////////////////////////////////////////////

// For 'EPS Surprise'
window.charts.epsSurpriseTooltip = function(values, chart) {
    var thisYear = values.date.getFullYear();
    var nextYear = thisYear + 1;
    if (chart.current.lines["consensus_" + thisYear]) {
        var currentYearConsensus = chart.getValue(values.date, "consensus_" + thisYear);
        var currentYearColor = chart.current.lines["consensus_" + thisYear].getAttribute("stroke");
        var currentYearString = "<br><span style='color:" + currentYearColor + "'>" + thisYear + " Consensus: " + currentYearConsensus.toFixed(2) + "</span>"
    } else {
        var currentYearString = "";
    }
    if (chart.current.lines["consensus_" + nextYear]) {
        nextYearDate = new Date(values.date.getTime());
        nextYearDate.setFullYear(nextYear);
        var nextYearConsensus = chart.getValue(nextYearDate, "consensus_" + nextYear);
        var nextYearColor = chart.current.lines["consensus_" + nextYear].getAttribute("stroke");
        var nextYearString = "<br><span style='color:" + nextYearColor + "'>" + nextYear + " Consensus: " + nextYearConsensus.toFixed(2) + "</span>"
    } else {
        var nextYearString = "";
    }
    return "Date: " + values.date.toDateString() + "<br>" +
        "EPS Surprise: " + values.value.toFixed(2) + "%" +
        currentYearString +
        nextYearString;
};

// For 'Monthly EPS'
window.charts.monthlyEpsTooltip = function(values, chart) {
    return "Date: " + values.date.toDateString() + "<br>" +
        "EPS: " + values.value.toFixed(2);
};

// For 'Broker Recommendation'
window.charts.brokerFormatter = function(values, chart) {
    return "Date: " + values.date.toDateString() + "<br>" +
        "Broker Recommendation: " + values.value.toFixed(2);
}

// For Dollar Values
window.charts.genericTooltipDollar = function(values, chart) {
    return 'Date: ' + values.date.toDateString() + '<br>' +
        'Value: ' + Number(values.value).toLocaleString('en-US', { style: 'currency', currency: 'USD', minimumFractionDigits: 2, maximumFractionDigits: 2 });
};

// For Percent Values
window.charts.genericTooltipPercent = function(values, chart) {
    return 'Date: ' + values.date.toDateString() + '<br>' +
        'Value: ' + Number(values.value).toLocaleString() + '%';
};

// For Number Values
window.charts.genericTooltipNumber = function(values, chart) {
    return 'Date: ' + values.date.toDateString() + '<br>' +
        'Value: ' + Number(values.value);
};


////////////////////////////////////////////////////////////////////////////////
/// CHART HOVER FORMATTERS
////////////////////////////////////////////////////////////////////////////////

// Dollar Formatting
window.charts.chartDataFormatterDollar = function(date, value) {
    return '<span class="symbol">' + this.current.symbol + '</span>' +
        '<span class="date">' + date.toDateString() + '</span>' +
        '<span class="value">' + Number(value).toLocaleString('en-US', { style: 'currency', currency: 'USD', minimumFractionDigits: 2, maximumFractionDigits: 2 }) + '</span>';
}

// Percent Formatting
window.charts.chartDataFormatterPercent = function(date, value) {
    return '<span class="symbol">' + this.current.symbol + '</span>' +
        '<span class="date">' + date.toDateString() + '</span>' +
        '<span class="value">' + Number(value).toLocaleString() + '%</span>';
}

// Number Formatting
window.charts.chartDataFormatterNumber = function(date, value) {
    return '<span class="symbol">' + this.current.symbol + '</span>' +
        '<span class="date">' + date.toDateString() + '</span>' +
        '<span class="value">' + Number(value) + '</span>';
}


////////////////////////////////////////////////////////////////////////////////
/// WRAPPER DEFINITIONS
////////////////////////////////////////////////////////////////////////////////

window.charts.wrappers = {
    'null': {
        key: 'null',
        api_key: 'null'
    },
    'price_and_consensus': {
        key: 'price',
        api_key: 'price_and_consensus',
        label: 'Price ($)',
        periodicity: true,
        overlays: [
            { key: 'consensus_2014', label: '2014' }, { key: 'consensus_2015', label: '2015' }, { key: 'consensus_2016', label: '2016' }, { key: 'consensus_2017', label: '2017' }, { key: 'consensus_2018', label: '2018' }
        ],
        legend_label: 'Consensus',
        has_volume: true
    },
    'price_and_eps_estimates_consensus': {
        key: 'price',
        api_key: 'price_and_eps_estimates_consensus',
        label: 'Price ($)',
        periodicity: true,
        annotation: window.charts.drawSurpriseAnnotations,
        annotation_key: 'eps_surprise',
        annotation_label: 'EPS Surprise',
        annotation_legend_values: [15, -15],
        annotation_formatter: window.charts.epsSurpriseTooltip,

        // if chart will not come check here
        overlays: [
             { key: 'consensus_2015', label: '2015' }, { key: 'consensus_2016', label: '2016' }, { key: 'consensus_2017', label: '2017' }, { key: 'consensus_2018', label: '2018' }
        ],
        overlay_label: 'EPS ($/sh)',
        legend_label: 'Consensus',
        has_volume: true
    },
    'price_and_eps_surprise': {
        key: 'price',
        api_key: 'price_and_eps_surprise',
        label: 'Price ($)',
        periodicity: true,
        annotation: window.charts.drawSurpriseAnnotations,
        annotation_key: 'eps_surprise',
        annotation_label: 'EPS Surprise',
        annotation_legend_values: [15, -15],
        annotation_formatter: window.charts.epsSurpriseTooltip,
        overlays: [],
        overlay_label: 'EPS ($/sh)',
        has_volume: true
    },
    '12_month_eps': {
        key: 'price',
        api_key: '12_month_eps',
        label: 'Price ($)',
        periodicity: true,
        overlays: [{
            key: 'monthly_consensus_est',
            label: 'Consensus Estimate',
            annotation: window.charts.drawCircle,
            annotation_key: 'monthly_consensus_est',
            annotation_formatter: window.charts.monthlyEpsTooltip,
            color: 'green'
        }, {
            key: '12_month_eps',
            label: '12 Month EPS',
            annotation: window.charts.drawCircle,
            annotation_key: '12_month_eps',
            annotation_formatter: window.charts.monthlyEpsTooltip,
            color: 'blue'
        }],
        overlay_label: 'EPS',
        has_volume: true
    },
    'broker_recommendations': {
        key: 'price',
        api_key: 'broker_recommendations',
        label: 'Price ($)',
        periodicity: true,
        overlays: [{
            key: 'broker_recommendation',
            label: 'Broker Recommendations',
            annotation: window.charts.drawX,
            annotation_key: 'broker_recommendation',
            annotation_formatter: window.charts.brokerFormatter,
            color: 'green',
            inverted_axis: true
        }],
        has_volume: true
    }

    ,
    'debt_to_capital': {
        key: 'debt_to_capital',
        api_key: 'debt_to_capital',
        label: 'Debt to Capital',
        periodicity: true,
        valid_if: function(value) { return value > 0; }
    },
    'pe_using_current_fy_eps_estimates': {
        key: 'pe_using_current_fy_eps_estimates',
        api_key: 'pe_using_current_fy_eps_estimates',
        label: 'Price to Earnings',
        periodicity: true
    },
    'price_to_book_value': {
        key: 'price_to_book_value',
        api_key: 'price_to_book_value',
        label: 'Price to Book Value',
        periodicity: true
    },
    'price_to_cash_flow': {
        key: 'price_to_cash_flow',
        api_key: 'price_to_cash_flow',
        label: 'Price to Cash Flow',
        periodicity: true
    },
    'price_to_sales': {
        key: 'price_to_sales',
        api_key: 'price_to_sales',
        label: 'Price to Sales',
        periodicity: true
    },
    'volume': {
        key: 'volume',
        api_key: 'volume',
        label: 'Volume',
        periodicity: true
    },
    'eps': {
        key: 'eps',
        api_key: 'eps',
        label: 'EPS',
        periodicity: true
    },
    'roe': {
        key: 'roe',
        api_key: 'roe',
        label: 'ROE',
        periodicity: true
    },
    'price_to_industry': {
        key: 'price',
        api_key: 'price_to_industry',
        label: 'Price',
        periodicity: true,
        overlays: [
            { key: 'industry_price', label: 'Industry Price' }
        ],
        overlay_label: 'Industry'
    },
    'sales_and_estimates': {
        key: 'sales',
        api_key: 'sales_and_estimates',
        label: 'Sales History & Est.',
        periodicity: true
    },
    'eps_and_estimates': {
        key: 'eps',
        api_key: 'eps_and_estimates',
        label: 'EPS History & Est.',
        periodicity: true
    },
    'industry_price': {
        key: 'x_industry_price',
        api_key: 'x_industry_price',
        label: 'Industry Price',
        periodicity: true
    },
    'eps_hist_and_est': {
        key: 'eps_hist_and_est_ttm_annual',
        api_key: 'eps_hist_and_est_ttm_annual',
        label: 'EPS Hist. and Est.',
        periodicity: false,
        labels_disable: false
    },
    'sales_hist_and_est': {
        key: 'sales_hist_and_est_ttm_annual',
        api_key: 'sales_hist_and_est_ttm_annual',
        label: 'Sales Hist. and Est.',
        periodicity: false,
        labels_disable: false
    },
    'beta': {
        key: 'beta',
        api_key: 'beta',
        label: 'Beta',
        seo_key: 'beta',
        annotation_key: 'beta',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipNumber,
        data_formatter: window.charts.chartDataFormatterNumber,
        periodicity: true
    },
    'book_value_per_share': {
        key: 'book_value_per_share',
        api_key: 'book_value_per_share',
        label: 'Book Value (Per Share)',
        seo_key: 'book-value',
        annotation_key: 'book_value_per_share',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: true
    },
    'cash': {
        key: 'cash',
        api_key: 'cash',
        seo_key: 'cash-equivalents-quarterly',
        label: 'Cash and Equivalents (Quarterly)',
        annotation_key: 'cash',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: true
    },
    'cash_from_financing': {
        key: 'cash_from_financing',
        api_key: 'cash_from_financing',
        seo_key: 'cash-from-financing-quarterly',
        label: 'Cash from Financing (Quarterly)',
        annotation_key: 'cash_from_financing',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: false
    },
    'cash_from_financing_ttm': {
        key: 'cash_from_financing_ttm',
        api_key: 'cash_from_financing_ttm',
        seo_key: 'cash-from-financing-ttm',
        label: 'Cash from Financing (TTM)',
        annotation_key: 'cash_from_financing_ttm',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: false
    },
    'cash_from_investing': {
        key: 'cash_from_investing',
        api_key: 'cash_from_investing',
        seo_key: 'cash-from-investing-quarterly',
        label: 'Cash from Investing (Quarterly)',
        annotation_key: 'cash_from_investing',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: false
    },
    'cash_from_investing_ttm': {
        key: 'cash_from_investing_ttm',
        api_key: 'cash_from_investing_ttm',
        seo_key: 'cash-from-investing-ttm',
        label: 'Cash from Investing (TTM)',
        annotation_key: 'cash_from_investing_ttm',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: false
    },
    'cash_from_operations': {
        key: 'cash_from_operations',
        api_key: 'cash_from_operations',
        seo_key: 'cash-from-operations-quarterly',
        label: 'Cash from Operations (Quarterly)',
        annotation_key: 'cash_from_operations',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: false
    },
    'cash_from_operations_ttm': {
        key: 'cash_from_operations_ttm',
        api_key: 'cash_from_operations_ttm',
        seo_key: 'cash-from-operations-ttm',
        label: 'Cash from Operations (TTM)',
        annotation_key: 'cash_from_operations_ttm',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: false
    },
    'debt_to_equity': {
        key: 'debt_to_equity',
        api_key: 'debt_to_equity',
        seo_key: 'debt-equity-ratio-quarterly',
        label: 'Debt to Equity Ratio (Quarterly)',
        annotation_key: 'debt_to_equity',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipPercent,
        data_formatter: window.charts.chartDataFormatterPercent,
        periodicity: true
    },
    'dividend': {
        key: 'dividend',
        api_key: 'dividend',
        seo_key: 'dividend-ttm',
        label: 'Dividend',
        annotation_key: 'dividend',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: true
    },
    'dividend_yield': {
        key: 'dividend_yield',
        api_key: 'dividend_yield',
        seo_key: 'dividend-yield-ttm',
        label: 'Dividend Yield (TTM)',
        annotation_key: 'dividend_yield',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipPercent,
        data_formatter: window.charts.chartDataFormatterPercent,
        periodicity: true
    },
    'eps_diluted': {
        key: 'eps_diluted',
        api_key: 'eps_diluted',
        seo_key: 'eps-diluted-quarterly',
        label: 'EPS Diluted (Quarterly)',
        annotation_key: 'eps_diluted',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: false
    },
    'eps_diluted_ttm': {
        key: 'eps_diluted_ttm',
        api_key: 'eps_diluted_ttm',
        seo_key: 'eps-diluted-ttm',
        label: 'EPS Diluted (TTM)',
        annotation_key: 'eps_diluted_ttm',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: false
    },
    'earnings_yield': {
        key: 'earnings_yield',
        api_key: 'earnings_yield',
        seo_key: 'earnings-yield-ttm',
        label: 'Earnings Yield (TTM)',
        annotation_key: 'earnings_yield',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipPercent,
        data_formatter: window.charts.chartDataFormatterPercent,
        periodicity: true
    },
    'enterprise_value': {
        key: 'enterprise_value',
        api_key: 'enterprise_value',
        seo_key: 'enterprise-value',
        label: 'Enterprise Value',
        annotation_key: 'enterprise_value',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: true
    },
    'free_cash_flow': {
        key: 'free_cash_flow',
        api_key: 'free_cash_flow',
        seo_key: 'free-cash-flow-quarterly',
        label: 'Free Cash Flow (Quarterly)',
        annotation_key: 'free_cash_flow',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: false
    },
    'free_cash_flow_ttm': {
        key: 'free_cash_flow_ttm',
        api_key: 'free_cash_flow_ttm',
        seo_key: 'free-cash-flow-ttm',
        label: 'Free Cash Flow (TTM)',
        annotation_key: 'free_cash_flow_ttm',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: true
    },
    'gross_profit_ttm': {
        key: 'gross_profit_ttm',
        api_key: 'gross_profit_ttm',
        seo_key: 'gross-profit-ttm',
        label: 'Gross Profit (TTM)',
        annotation_key: 'gross_profit_ttm',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: false
    },
    'gross_profit_margin': {
        key: 'gross_profit_margin',
        api_key: 'gross_profit_margin',
        seo_key: 'gross-margin-ttm',
        label: 'Gross Profit Margin (Quarterly)',
        annotation_key: 'gross_profit_margin',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipPercent,
        data_formatter: window.charts.chartDataFormatterPercent,
        periodicity: true
    },
    'market_cap': {
        key: 'market_cap',
        api_key: 'market_cap',
        seo_key: 'market-cap',
        label: 'Market Cap',
        annotation_key: 'market_cap',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: true
    },
    'net_income': {
        key: 'net_income',
        api_key: 'net_income',
        seo_key: 'net-income-quarterly',
        label: 'Net Income (Quarterly)',
        annotation_key: 'net_income',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: false
    },
    'net_income_ttm': {
        key: 'net_income_ttm',
        api_key: 'net_income_ttm',
        seo_key: 'net-income-ttm',
        label: 'Net Income (TTM)',
        annotation_key: 'net_income_ttm',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: true
    },
    'pe_ratio': {
        key: 'pe_ratio',
        api_key: 'pe_ratio',
        seo_key: 'pe-ratio-ttm',
        label: 'PE Ratio (TTM)',
        annotation_key: 'pe_ratio',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipNumber,
        data_formatter: window.charts.chartDataFormatterNumber,
        periodicity: true
    },
    'peg_ratio': {
        key: 'peg_ratio',
        api_key: 'peg_ratio',
        seo_key: 'peg-ratio-ttm',
        label: 'PEG Ratio (TTM)',
        annotation_key: 'peg_ratio',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipNumber,
        data_formatter: window.charts.chartDataFormatterNumber,
        periodicity: true
    },
    'ps_ratio': {
        key: 'ps_ratio',
        api_key: 'ps_ratio',
        seo_key: 'ps-ratio-ttm',
        label: 'PS Ratio (TTM)',
        annotation_key: 'ps_ratio',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipNumber,
        data_formatter: window.charts.chartDataFormatterNumber,
        periodicity: true
    },
    'price': {
        key: 'price',
        api_key: 'price',
        seo_key: 'price',
        label: 'Price',
        annotation_key: 'price',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: true
    },
    'price_to_book_value': {
        key: 'price_to_book_value',
        api_key: 'price_to_book_value',
        seo_key: 'price-book-value',
        label: 'Price to Book Value',
        annotation_key: 'price_to_book_value',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: true
    },
    'q_profit_margin': {
        key: 'q_profit_margin',
        api_key: 'q_profit_margin',
        seo_key: 'profit-margin-quarterly',
        label: 'Profit Margin (Quarterly)',
        annotation_key: 'q_profit_margin',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipPercent,
        data_formatter: window.charts.chartDataFormatterPercent,
        periodicity: true
    },
    'return_on_assets': {
        key: 'return_on_assets',
        api_key: 'return_on_assets',
        seo_key: 'return-on-assets-ttm',
        label: 'Return on Assets (TTM)',
        annotation_key: 'return_on_assets',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipPercent,
        data_formatter: window.charts.chartDataFormatterPercent,
        periodicity: true
    },
    'roe': {
        key: 'roe',
        api_key: 'roe',
        seo_key: 'return-on-equity-ttm',
        label: 'Return on Equity (TTM)',
        annotation_key: 'roe',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipPercent,
        data_formatter: window.charts.chartDataFormatterPercent,
        periodicity: true
    },
    'revenue_yoy': {
        key: 'revenue_yoy',
        api_key: 'revenue_yoy',
        seo_key: 'revenue-quarterly-yoy-growth',
        label: 'Revenue (Quarterly YoY Growth)',
        annotation_key: 'revenue_yoy',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipPercent,
        data_formatter: window.charts.chartDataFormatterPercent,
        periodicity: false
    },
    'revenue': {
        key: 'revenue',
        api_key: 'revenue',
        seo_key: 'revenue-quarterly',
        label: 'Revenue (Quarterly)',
        annotation_key: 'revenue',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: false
    },
    'revenue_ttm': {
        key: 'revenue_ttm',
        api_key: 'revenue_ttm',
        seo_key: 'revenue-ttm',
        label: 'Revenue (TTM)',
        annotation_key: 'revenue_ttm',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: true
    },
    'share_holders_equity': {
        key: 'share_holders_equity',
        api_key: 'share_holders_equity',
        seo_key: 'shareholders-equity-quarterly',
        label: 'Shareholders Equity (Quarterly)',
        annotation_key: 'share_holders_equity',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: false
    },
    'assets_total': {
        key: 'assets_total',
        api_key: 'assets_total',
        seo_key: 'total-assets-quarterly',
        label: 'Total Assets (Quarterly)',
        annotation_key: 'assets_total',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: false
    },
    'expenses_total': {
        key: 'expenses_total',
        api_key: 'expenses_total',
        seo_key: 'total-expenses-quarterly',
        label: 'Total Expenses (Quarterly)',
        annotation_key: 'expenses_total',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: true
    },
    'expenses_total_ttm': {
        key: 'expenses_total_ttm',
        api_key: 'expenses_total_ttm',
        seo_key: 'total-expenses-ttm',
        label: 'Total Expenses (TTM)',
        annotation_key: 'expenses_total_ttm',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: true
    },
    'liabilities_total': {
        key: 'liabilities_total',
        api_key: 'liabilities_total',
        seo_key: 'total-liabilities-quarterly',
        label: 'Total Liabilities (Quarterly)',
        annotation_key: 'liabilities_total',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: false
    },
    'debt_lt_total_ttm': {
        key: 'debt_lt_total_ttm',
        api_key: 'debt_lt_total_ttm',
        seo_key: 'total-long-term-debt-quarterly',
        label: 'Total Long Term Debt (Quarterly)',
        annotation_key: 'debt_lt_total_ttm',
        annotation: window.charts.drawAnnotaionFundamental1,
        annotation_formatter: window.charts.genericTooltipDollar,
        periodicity: false
    },
    'sector_price_index': {
        key: 'sector_price_index',
        api_key: 'sector_price_index_909',
        label: 'Sector Price Index',
        periodicity: false,
        overlays: [{ key: 'ivv_price', label: 'IVV Price', color: 'red' }],
        overlay_label: 'IVV'
    },
    'industry_price_index': {
        key: 'industry_price_index',
        api_key: 'industry_price_index_909',
        label: 'Industry Price Index',
        periodicity: false,
        overlays: [{ key: 'ivv_price', label: 'IVV Price', color: 'red' }],
        overlay_label: 'IVV'
    },
    'sector_return_perc': {
        key: 'sector_return_perc',
        api_key: 'sector_return_perc',
        label: 'Total Return %',
        overlays: [{ key: 'ivv_return_perc', label: 'IVV Total Return %', color: 'black', layer: 0, show_label: true }],
        overlay_label: 'IVV Total Return %',
        data_formatter: window.charts.chartDataFormatterPercent,
        periodicity: true
    },
    'industry_return_perc': {
        key: 'industry_return_perc',
        api_key: 'industry_return_perc',
        label: 'Total Return %',
        overlays: [{ key: 'ivv_return_perc', label: 'IVV Total Return %', color: 'black', layer: 0, show_label: true }],
        overlay_label: 'IVV Total Return %',
        data_formatter: window.charts.chartDataFormatterPercent,
        periodicity: true
    },
    'sector_return_perc_1yr': {
        key: 'sector_return_perc',
        api_key: 'sector_return_perc_1yr',
        label: 'Total Return %',
        overlays: [{ key: 'ivv_return_perc', label: 'IVV Total Return %', color: 'black', layer: 0, show_label: true }],
        overlay_label: 'IVV Total Return %',
        data_formatter: window.charts.chartDataFormatterPercent,
        periodicity: false
    },
    'industry_return_perc_1yr': {
        key: 'industry_return_perc',
        api_key: 'industry_return_perc_1yr',
        label: 'Total Return %',
        overlays: [{ key: 'ivv_return_perc', label: 'IVV Total Return %', color: 'black', layer: 0, show_label: true }],
        overlay_label: 'IVV Total Return %',
        data_formatter: window.charts.chartDataFormatterPercent,
        periodicity: false
    }
};


////////////////////////////////////////////////////////////////////////////////
/// MISC FUNCTIONS
////////////////////////////////////////////////////////////////////////////////


// Format text date value into object
window.charts.get_date_object = function(val) {
    if (val) {
        val = val.split('/');

        // Handle either 2-digit or 4-digit year
        if (val[2].length == 2)
            var year = parseInt(val[2]) > 50 ? '19' + val[2] : '20' + val[2];
        else
            var year = val[2];

        var month = val[0] - 1;
        var day = val[1];

        val = new Date(year, month, day);
    }

    return val;
}

// Format date object into text string
window.charts.get_date_string = function(d, format_str) {
    var out = '';

    if (d) {
        var day = d.getDate();
        var month = d.getMonth() + 1;
        var year4 = d.getFullYear();
        var year2 = year4.toString().substr(2, 2);

        if (day < 10)
            day = '0' + day.toString();

        if (month < 10)
            month = '0' + month.toString();

        out = format_str.replace('dd', day).replace('mm', month).replace('yyyy', year4).replace('yy', year2);
    }

    return out;
}

// Initialize independant overlays (e.g. Fundamental Charts)
window.charts.init_independant_overlays = function(form_obj, price_chart) {
    var _overlay_colors = form_obj.find('.overlay_colors').val() ? form_obj.find('.overlay_colors').val().split(',') : window.charts.default_overlay_colors;
    var _annotations = form_obj.find('.overlay_annotations').val() ? form_obj.find('.overlay_annotations').val().split(',') : false;
    var _right_col_price = form_obj.find('.right_col_price').val() ? form_obj.find('.right_col_price').val() : true;

    // Initialize onchange event
    $('.independant_overlays').on('change', function(e) {
        var obj = $(this);
        var _selected_value = false;
        var _layer = parseInt($(this).attr('data-charts-layer'), 10);

        if ($(this).is('select')) {
            $(this).find('option[value=' + $(this).val() + ']').attr('selected', 'selected');
        }

        if ($(this).is('select')) {
            _selected_value = $(obj).find('option[selected]').val();
        } else if ($(this).is('input[type=checkbox]') && $(this).is(':checked')) {
            _selected_value = $(obj).val();
        } else if ($(this).is('input[type=checkbox]') && !$(this).is(':checked')) {
            _selected_value = $(obj).val() * -1;
        }

        // Only set overlay if option is selected
        if (_selected_value !== false) {
            var _wrapper;
            var _label;
            var _annotation_formatter;
            var _overlay_color;
            var _type = ($(obj).attr('rel') !== undefined) ? $(obj).attr('rel') : false;
            var _symbol = form_obj.find('.symbol').val();

            // Enable 'dummy' overlay in layer 1 to prevent bugs from appearing
            // (If there is actual layer 1 overlay, it will be enabled below)
            //
            // JLP 11/11/16 - Commented out, appears to no longer be necessary with revised layer 
            //                logic and caused new bug to appear with overlays defined in wrapper.
            //
            // price_chart.enableOverlay( { key: 'null', layer: 1 }); 

            // Overlay type defined by 'rel' attribute, for example Moving Averages
            if (_type == 'ma') {
                // If _layer is not defined, default to 2
                if (isNaN(_layer))
                    _layer = 2;

                switch (_selected_value) {
                    case '50':
                        _overlay_color = window.charts.default_ma_colors[1];
                        break;

                    case '200':
                        _overlay_color = window.charts.default_ma_colors[2];
                        break;

                    default:
                        _overlay_color = window.charts.default_ma_colors[0];
                        break;
                }

                if (_selected_value > 0) {
                    // Enable MA
                    price_chart.enableOverlay({
                        key: 'price',
                        label: _selected_value + ' Day MA',
                        symbol: _symbol,
                        type: 'computed_moving_average',
                        moving_average_days: _selected_value,
                        color: _overlay_color,
                        layer: _layer
                    });
                } else {
                    // Switch _selected_value back to positive
                    _selected_value = _selected_value * -1;

                    // Disable overlay for this MA
                    price_chart.disableOverlay({
                        key: 'price',
                        type: 'computed_moving_average',
                        symbol: _symbol,
                        moving_average_days: _selected_value
                    });
                }
            } else {
                // Disable overlay for just the current layer
                price_chart.disableOverlay({ layer: _layer });

                // Overlay type inferred by primary focus of chart -- the wrapper or the symbol
                if (!_type) {
                    // Are we getting a wrapper or symbol from the dropdown?
                    if (form_obj.find('.wrapper') && form_obj.find('.wrapper').attr('type')) {
                        // Form has wrapper element, dropdown passed a symbol
                        _wrapper = form_obj.find('.wrapper').val();
                        _symbol = $(obj).val();
                        _label = $(obj).find('option:selected').text();
                    } else if (form_obj.find('.symbol') && form_obj.find('.symbol').attr('type')) {
                        // Form has symbol element, dropdown passed a wrapper
                        _wrapper = $(obj).val();
                        _label = window.charts.get_wrapper_title(_wrapper);
                    }
                }

                _annotation_formatter = window.charts.wrappers[_wrapper]['annotation_formatter'];

                if (_symbol && _wrapper) {
                    // Get the index
                    var _index = _layer - 2;

                    // Get the label
                    var _color = (_overlay_colors[_index]) ? _overlay_colors[_index] : _overlay_colors[0];

                    price_chart.enableOverlay({
                        key: _wrapper,
                        label: _label,
                        layer: _layer,
                        symbol: _symbol,
                        color: _color,
                        annotation: (_annotations[_index]) ? window.charts[_annotations[_index]] : null,
                        annotation_key: (_annotations[_index]) ? _wrapper : null,
                        annotation_formatter: (_annotation_formatter) ? _annotation_formatter : null
                    });

                    if (price_chart.label_container) {
                        window.charts.set_external_chart_label('.' + price_chart.label_container, 'dyn_' + _layer, _color, _label);
                    } else {
                        // Set overlay label inside chart
                        price_chart.set('overlay_label', _label, false);
                    }
                }
            }
        }

        price_chart.update();
    });
}

// If element with class of 'independant_overlays' is checked, trigger the 'change' event
window.charts.trigger_checked_overlays = function() {
    var independant_overlays_arr = $('.independant_overlays');

    $.each(independant_overlays_arr, function(idx, val) {
        if ($(val).is(':checked') || ($(val).find('option[selected=selected]')[0])) {
            $(val).trigger('change');
        }
    });
}

// Return the wrapper title, either from definitions or programatically
window.charts.get_wrapper_title = function(overlay) {
    var _title;

    if (window.charts.wrappers[overlay] && window.charts.wrappers[overlay]['label']) {
        _title = window.charts.wrappers[overlay]['label'];
    } else {
        _title = overlay.replace(/_/g, ' ').toTitleCase();
    }

    return _title;
}

// Resize chart when browser window is too small
window.charts.setSize = function(form_obj) {

    var w = $(window).width();
    var _width, _height;

    if (w) {
        _width = parseInt(form_obj.find('.width').val(), 10);
        _height = parseInt(form_obj.find('.height').val(), 10);
        /*if( w < 1024 )
        {
            _width  = (w - 100);
            _height = Math.round(_width / 2.05);
        }
        else
        {
            _width  = parseInt( form_obj.find('.width').val(), 10 );
            _height = parseInt( form_obj.find('.height').val(), 10 );
        }*/

    }

    return [{ "windowsize": w, "width": _width, "height": _height }];
}

// Set external chart label
window.charts.set_external_chart_label = function(_container, _element, _color, _label) {
    if ($(_container).find('p.' + _element)[0]) {
        // Update existing element
        $(_container).find('p.' + _element).html('<span style="background:' + _color + '">&nbsp;</span>' + _label);
    } else {
        // Create new element
        $(_container).append('<p class="' + _element + '"><span style="background:' + _color + '">&nbsp;</span>' + _label + '</p>');
    }
};

// Initialize a chart
window.charts.init_charts = function(form_id, chart_id) {
    var form_obj = $('#' + form_id);
    var chart_obj = $('#' + chart_id);

    // Get chart dimensions in a responsive manner
    var chart_dims = window.charts.setSize(form_obj);

    if (form_obj[0] && chart_obj[0]) {
        var _width = chart_dims[0].width;
        var _height = chart_dims[0].height;
        var _symbol = form_obj.find('.symbol').val();
        var _wrapper_sel = form_obj.find('.wrapper :selected').val();
        var _wrapper_hid = form_obj.find('.wrapper').val();
        var _period_sel = form_obj.find('.period :selected').val();
        var _period_hid = form_obj.find('.period').val();
        var _period = (_period_hid) ? _period_hid : _period_sel; // Prefer hidden form fields
        var _default_period = (form_obj.find('.default_period').val()) ? form_obj.find('.default_period').val() : null;
        var _chart_type = form_obj.find('.chart_type').val() ? form_obj.find('.chart_type').val() : 'line';
        var _sec_type = (form_obj.find('.secondary_type').val()) ? form_obj.find('.secondary_type').val() : null;
        var _sec_wrapper = (form_obj.find('.secondary_wrapper').val()) ? form_obj.find('.secondary_wrapper').val() : null;
        var _annotations = form_obj.find('.annotations').val() ? true : false;
        var _overlays = form_obj.find('.overlays').val() ? true : false;
        var _foreground_colors = form_obj.find('.colors').val() ? form_obj.find('.colors').val().split(',') : window.charts.default_colors;
        var _background_color = form_obj.find('.background_color').val() ? form_obj.find('.background_color').val().split(',') : ['#E6E6E6', '#FEFEFE'];
        var _secondary_background_color = form_obj.find('.secondary_background_color').val() ? form_obj.find('.secondary_background_color').val() : 'transparent';
        var _bar_color = form_obj.find('.bar_color').val() ? form_obj.find('.bar_color').val() : '#4f95cc';
        var _secondary_height = form_obj.find('.secondary_height').val() ? form_obj.find('.secondary_height').val() : 0.3;
        var _default_stroke = form_obj.find('.default_stroke').val() ? form_obj.find('.default_stroke').val() : 1;
        var _ymin = form_obj.find('.ymin').val() ? parseInt(form_obj.find('.ymin').val(), 10) : null;
        var _ymax = form_obj.find('.ymax').val() ? parseInt(form_obj.find('.ymax').val(), 10) : null;
        var _ymin_overlay = form_obj.find('.ymin_overlay').val() ? parseInt(form_obj.find('.ymin_overlay').val(), 10) : null;
        var _ymax_overlay = form_obj.find('.ymax_overlay').val() ? parseInt(form_obj.find('.ymax_overlay').val(), 10) : null;
        var _show_legend = (form_obj.find('.show_legend').val() && form_obj.find('.show_legend').val() === 'false') ? false : true;
        var _from_date = window.charts.get_date_object(form_obj.find('.from_date').val());
        var _to_date = window.charts.get_date_object(form_obj.find('.to_date').val());
        /*var _data_source                = HTTP_PROTOCOL_JS + WIDGET3_PATH_JS + '/data/chart/json/{SYMBOL}/{WRAPPER}/' + window.location.hostname;*/
        var _data_source = form_obj.find('.source_id').val() + '/data/chart/json/{SYMBOL}/{WRAPPER}/' + form_obj.find('.site_root_path').val();
        var _show_grid = form_obj.find('.show_grid').val() ? form_obj.find('.show_grid').val() : true;
        var _margins = form_obj.find('.margins').val() ? form_obj.find('.margins').val().split(',') : [30, 60, 30, 60];
        var _label_container = form_obj.find('.external_labels').val() ? form_obj.find('.external_labels').val().split(',') : false;
        var _wrapper_table = form_obj.find('.wrapper_table').val() ? form_obj.find('.wrapper_table').val().split(',') : false;
        var _mode = form_obj.find('.mode').val() ? form_obj.find('.mode').val() : false;
        var _draw_mode_default_width = form_obj.find('.draw_mode_default_width').val() ? form_obj.find('..draw_mode_default_width').val() : 2;
        var _price_info_on_hover = form_obj.find('.price_info_on_hover').val() ? form_obj.find('.price_info_on_hover').val() : true;
        var _enable_zoom = (form_obj.find('.enable_zoom').val() && form_obj.find('.enable_zoom').val() === 'false') ? false : true;
        var _end_values = (form_obj.find('.end_values').val() && form_obj.find('.end_values').val() === 'true') ? true : false;
        var _end_value_style = form_obj.find('.end_value_style').val() ?
            'PriceChart.Config.' + form_obj.find('.end_value_style').val() :
            'PriceChart.Config.END_VALUE_STYLE__VALUES';
        var _end_value_function = form_obj.find('.end_value_function').val() ? window.charts[form_obj.find('.end_value_function').val()] :
            false;

        // Prefer hidden form fields; otherwise look at drop downs
        var _wrapper = (_wrapper_hid) ? _wrapper_hid : _wrapper_sel;

        // Set initial values
        var price_chart = new PriceChart({
            'data_source': _data_source,
            'width': _width,
            'height': _height,
            'secondary_height': _secondary_height,
            'margins': { 'top': parseInt(_margins[0], 10), 'right': parseInt(_margins[1], 10), 'bottom': parseInt(_margins[2], 10), 'left': parseInt(_margins[3], 10) },
            'canvas': '#' + chart_id,
            'compatibility_mode': typeof(_price_chart_compatibility_mode) != 'undefined' ? true : false,
            'wrappers': window.charts.wrappers,
            'png_convert_endpoint': '/charts/svg_to_png.php',
            'data_formatter': window.charts.chartDataFormatterDollar,
            'transition_duration': 0,
            'background_color': _background_color,
            'secondary_background_color': _secondary_background_color,
            'bar_color': _bar_color,
            'bar_spacing': 0.5,
            'colors': _foreground_colors,
            'draw_mode_colors': ['orange', 'red', 'goldenrod', 'green', 'blue', 'purple'],
            'line_properties': {
                'base': { 'stroke-width': _default_stroke + 'px' },
                'default': { 'stroke-width': _default_stroke + 'px' },
                'monthly_consensus_est': { 'stroke-dasharray': '5,5', 'node': '3px' },
                '12_month_eps': { 'node': '3px' }
            },
            'initial_values': {
                'symbol': _symbol,
                'wrapper': _wrapper,
                'period': _period,
                'from_date': _from_date,
                'to_date': _to_date,
                'chart_type': _chart_type,
                'secondary_chart_type': _sec_type,
                'secondary_wrapper': _sec_wrapper,
                'annotations': _annotations,
                'overlays': _overlays
            },
            'watermark': {
                //'source': STATIC_CONTENT_ROOT_JS + '/images/zacks/chart/zacks-watermark-300x180-01-opacity-100.png'
                //'source': STATIC_CONTENT_ROOT_JS + '/images/zacks/chart/zacks-watermark-300x180-02-opacity-50.png'
                //'source': STATIC_CONTENT_ROOT_JS + '/images/zacks/chart/zacks-watermark_300x180_opacity-50.png'
                'source': form_obj.find('.img_path').val() + '/images/zacks/chart/zacks-watermark_300x180_opacity-50.png',
                'width': '120px',
                'height': '40px',
                'x': '0%',
                'y': '100%'
            },
            'error_image': form_obj.find('.img_path').val() + '/images/zacks/chart/zacks-watermark_300x180_opacity-50.png',
            'ymin': _ymin,
            'ymax': _ymax,
            'ymin_overlay': _ymin_overlay,
            'ymax_overlay': _ymax_overlay,
            'show_legend': _show_legend,
            'show_grid': _show_grid,
            'mode': _mode,
            'draw_mode_default_width': _draw_mode_default_width,
            'price_info_on_hover': _price_info_on_hover,
            'show_end_values': _end_values,
            'end_value_style': _end_value_style,
            'end_value_function': _end_value_function
        });

        // Set object for external properties
        price_chart.external = {};

        // Set object property for first event trigger
        price_chart.first_trigger = true;

        // Set object property for external label container
        price_chart.label_container = _label_container;

        // Set object property for wrapper table
        price_chart.wrapper_table = _wrapper_table;

        // Set object property for chart DOM ID
        price_chart.chart_domid = chart_id;

        // Set object property for array of chart tables
        price_chart.data_table = {};

        // Set object property to keep track of mode
        price_chart.external.mode = _mode;

        // If this chart uses an 'index wrapper', use value of selected key for chart label
        // Examples of index wrapper charts: sector index, industry index
        if (form_obj.find('.wrapper_type')[0] && (form_obj.find('.wrapper_type').val() == 'index')) {
            var _label = form_obj.find('.symbol option:selected').text();
            var _wrapper = form_obj.find('.wrapper').val();

            // Try alternate ways to get a label
            if (_label == '' || _label == undefined) {
                _label = (form_obj.find('.symbol_description')[0]) ? form_obj.find('.symbol_description').val() : form_obj.find('.symbol').val();
            }

            if (price_chart.label_container) {
                // Set external label
                window.charts.set_external_chart_label('.' + price_chart.label_container, 'dyn_symbol', _foreground_colors[0], _label);
            } else {
                // Update wrapper label
                window.charts.wrappers[_wrapper].label = _label;
            }
        }

        // Events which take place after chart rendering is complete
        price_chart.events.on('renderComplete', function() {
            if (price_chart.first_trigger === false) {
                price_chart.first_trigger = true;

                // Build wrapper table if option is enabled
                /*if(price_chart.wrapper_table)
                {
                    price_chart.set_wrapper_table();
                }*/

                // Enable overlays which might be enabled by default
                window.charts.trigger_checked_overlays();

                // Set default period of charts with period buttons
                if (_default_period) {
                    $(form_obj).find('[data-value=' + _default_period + ']').click();
                }
            }

            var symbol = $('#' + form_id + ' .symbol').val();

            if (symbol && symbol != '') {
                // Enable Chart Submit button
                $('#' + form_id + ' input[type=submit]').removeAttr('disabled');
            }

            // To hide loading text in quote price chart page
            $(".actHideLoading").hide();

        });

        // Call chart library
        price_chart.update();

        // Add to global array of charts to reference later
        window.charts.chart_list.push({ 'price_chart': price_chart, 'form_obj': form_obj, 'chart_obj': chart_obj });

        // Initialize independant overlays (e.g. Fundamental Charts)
        window.charts.init_independant_overlays(form_obj, price_chart);

        if (price_chart.error !== null) {
            console.log('Could not build chart: ' + price_chart.error);
        }

        // If no form action, cancel submit.  PNG export to browser.
        $(form_obj).on('submit', function(e) {
            if ($(form_obj).find('input[name="action"]').val() === undefined) {
                e.preventDefault();
            } else {
                // If there is a form action, that means we are in a chart building form.  PNG is saved on server.
                var xml = (new XMLSerializer).serializeToString($('#chart_canvas svg')[0]);
                $(form_obj).prepend('<input type="hidden" name="svg" value="' + encodeURIComponent(xml) + '" />');
            }
        });

        // Changed symbol
        $('#' + form_id + ' .symbol').on('change', function(e) {
            var _new_symbol;
            var _label;
            var _wrapper = form_obj.find('.wrapper').val();

            // Disable form submit while chart is loading
            $('#' + form_id + ' input[type=submit]').attr('disabled', 'disabled');

            if ($(this).is('select')) {
                // Industry/Sector pages have dropdown for symbol
                _new_symbol = $(this).find(':selected').val();
                _label = $(this).find(':selected').text();
            } else {
                // Tuners have text input element for symbol
                _new_symbol = $(this).val();
                _label = $(this).text();
            }

            if (price_chart.label_container) {
                // Set external label
                window.charts.set_external_chart_label('.' + price_chart.label_container, 'dyn_symbol', _foreground_colors[0], _label);
            } else {
                // Update wrapper label
                window.charts.wrappers[_wrapper].label = _label;
            }

            // Set new symbol
            price_chart.set('symbol', _new_symbol, false);

            price_chart.update();
        });

        // Changed wrapper
        $('#' + form_id + ' .wrapper').on('change', function(e) {
            if (!$(this).hasClass('noupdate')) {
                // Disable form submit while chart is loading
                $('#' + form_id + ' input[type=submit]').attr('disabled', 'disabled');

                var _new_wrapper = $(this).find(':selected').val();

                // Set new wrapper
                price_chart.set('wrapper', _new_wrapper, false);

                // Only show Volume for charts which have it defined in wrapper configuration
                if (window.charts.wrappers[_new_wrapper].has_volume && window.charts.wrappers[_new_wrapper].has_volume === true) {
                    $('.show-volume').prop('checked', true);
                    $('.secondary_type').val('bar');
                    $('.secondary_wrapper').val('volume');
                    price_chart.set('secondary_chart_type', 'bar', false);
                    price_chart.set('secondary_wrapper', 'volume', false);
                } else {
                    $('.show-volume').prop('checked', false);
                    $('.secondary_type').val('');
                    $('.secondary_wrapper').val('');
                    price_chart.set('secondary_chart_type', null, false);
                    price_chart.set('secondary_wrapper', null, false);
                }

                var today = new Date();
                var to_date, dd, mm, yy = '';

                if (_new_wrapper == '12_month_eps') {
                    // Set 'to_date' 3 years in the future for 12 Month EPS Chart
                    dd = today.getDate();
                    mm = today.getMonth() + 1; //January is 0!
                    yy = today.getFullYear() + 3 + '';
                    if (dd < 10) { dd = '0' + dd; }
                    if (mm < 10) { mm = '0' + mm; }
                    to_date = mm + '/' + dd + '/' + yy.substr(2, 2);
                } else {
                    // Set 'to_date' to current date
                    dd = today.getDate();
                    mm = today.getMonth() + 1; //January is 0!
                    yy = today.getFullYear() + '';
                    if (dd < 10) { dd = '0' + dd; }
                    if (mm < 10) { mm = '0' + mm; }
                    to_date = mm + '/' + dd + '/' + yy.substr(2, 2);
                }

                // Set 'to_date' value
                $('.to_date').val(to_date);
                price_chart.set('to_date', new Date(to_date), false);

                price_chart.update();

                // Enable overlays which might be enabled by default
                window.charts.trigger_checked_overlays();
            }
        });

        // Changed primary colors
        $('#' + form_id + ' .colors').on('change', function(e) {
            var _colors = $('#' + form_id + ' .colors').val();
            _colors = (_colors.indexOf(',') == -1) ? [_colors] : _colors.split(',');

            price_chart.redraw({ 'colors': _colors });
        });

        // Changed periodicity
        $('#' + form_id + ' .period').on('change', function(e) {
            price_chart.set('period', $(this).find(':selected').val());
        });

        // Changed 'from' date
        $('#' + form_id + ' .from_date').on('change', function(e) {
            // Clear period buttons
            $('#' + form_id + ' button.timeperiod').css('border-style', 'outset');
            price_chart.set('from_date', $(this).val());
        });

        // Changed 'to' date
        $('#' + form_id + ' .to_date').on('change', function(e) {
            // Clear period buttons
            $('#' + form_id + ' button.timeperiod').css('border-style', 'outset');
            price_chart.set('to_date', $(this).val());
        });

        // Changed time period
        $('#' + form_id + ' button.timeperiod').on('click', function(e) {
            // Disable form submit while chart is loading
            $('#' + form_id + ' input[type=submit]').attr('disabled', 'disabled');

            e.preventDefault();
            var today = new Date();
            var period_start = new Date();
            var periodicity;

            // Subtract one day from starting date
            period_start.setDate(period_start.getDate() - 1);

            // Clear period buttons & start/end dates
            form_obj.find('.from_date').val('');
            form_obj.find('.to_date').val('');
            $('#' + form_id + ' button.timeperiod').css('border-style', 'outset');

            // Indicate active button
            $(this).css('border-style', 'inset');

            switch ($(this).attr('data-value')) {
                // FOLLOWING CASES ARE USED BY TICKER CHARTS
                case '1mo-daily':
                    period_start.setMonth(today.getMonth() - 1);
                    periodicity = 'daily';
                    break;
                case '3mo-daily':
                    period_start.setMonth(today.getMonth() - 3);
                    periodicity = 'daily';
                    break;
                case 'ytd-daily':
                    // Requirement per JK, 1/3/2017: If current month is Jan, start date should be previous year
                    if (today.getMonth() == 0)
                        period_start.setFullYear(today.getFullYear() - 1, 0, 1);
                    else
                        period_start.setFullYear(today.getFullYear(), 0, 1);

                    periodicity = 'daily';
                    break;
                case '1yr-weekly':
                    period_start.setFullYear(today.getFullYear() - 1);
                    periodicity = 'weekly';
                    break;
                case '3yr-weekly':
                    period_start.setFullYear(today.getFullYear() - 3);
                    periodicity = 'weekly';
                    break;
                case '5yr-weekly':
                    period_start.setFullYear(today.getFullYear() - 5);
                    periodicity = 'weekly';
                    break;

                    // FOLLOWING CASES ARE USED BY INDUSTRY CHARTS
                case '1mo':
                    period_start.setMonth(today.getMonth() - 1);
                    periodicity = '1mo';
                    break;
                case '3mo':
                    period_start.setMonth(today.getMonth() - 3);
                    periodicity = '3mo';
                    break;
                case '6mo':
                    period_start.setMonth(today.getMonth() - 6);
                    periodicity = '6mo';
                    break;
                case 'ytd':
                    // Requirement per JK, 1/3/2017: If current month is Jan, start date should be previous year
                    if (today.getMonth() == 0)
                        period_start.setFullYear(today.getFullYear() - 1, 0, 1);
                    else
                        period_start.setFullYear(today.getFullYear(), 0, 1);

                    periodicity = 'ytd';
                    break;
                case '1yr':
                    period_start.setFullYear(today.getFullYear() - 1);
                    periodicity = '1yr';
                    break;
                case '3yr':
                    period_start.setFullYear(today.getFullYear() - 3);
                    periodicity = '3yr';
                    break;
                case '5yr':
                    period_start.setFullYear(today.getFullYear() - 5);
                    periodicity = '5yr';
                    break;
            };

            //console.log( 'start period: ' + new Date(period_start) );
            //console.log( 'end period: ' + new Date(today) );

            // Special consideration for industry/sector start date
            if (_wrapper == 'sector_return_perc' || _wrapper == 'industry_return_perc' ||
                _wrapper == 'sector_return_perc_1yr' || _wrapper == 'industry_return_perc_1yr'
            ) {
                // Subtract another day from starting date
                period_start.setDate(period_start.getDate() - 1);

                // If resulting day falls on a weekend, subtract additional days as necessary
                if (period_start.getDay() == 0) {
                    period_start.setDate(period_start.getDate() - 2); // Sunday
                } else if (period_start.getDay() == 6) {
                    period_start.setDate(period_start.getDate() - 1); // Saturday
                }
            }

            // Set From/To Date inputs
            $('#' + form_id + ' .from_date').val(window.charts.get_date_string(new Date(period_start), 'mm/dd/yy'));
            $('#' + form_id + ' .to_date').val(window.charts.get_date_string(new Date(today), 'mm/dd/yy'));

            // Set date range option
            price_chart.set('date_range', [period_start, today], false);
            price_chart.set('period', periodicity, false);
            price_chart.update();

            if (price_chart.error !== null)
                console.log('Could not build chart: ' + price_chart.error);
        });

        // Changed chart dimensions
        $('#' + form_id + ' .width,#' + form_id + ' .height').on('change', function() {
            _width = form_obj.find('.width').val();
            _height = form_obj.find('.height').val();

            // Redraw function seems to be bugged and causes PNG export to not work.
            //price_chart.redraw({ 'width':_width, 'height':_height});
        });

        // Init chart size selector
        // Used by zadmin chart tuner
        $('.chart-size').on('change', function() {
            var size_array = {
                1: [492, 240],
                2: [328, 160],
                3: [246, 120]
            }

            var _size = $(this).val();
            var _width = size_array[_size][0];
            var _height = size_array[_size][1];

            form_obj.find('.width').val(_width);
            form_obj.find('.height').val(_height);

            // Redraw function seems to be bugged and causes PNG export to not work.
            //price_chart.redraw({ 'width':_width, 'height':_height});

            price_chart.destroy();
            window.charts.init_charts(form_id, chart_id);
        });


        // Enable/Disable Overlays
        $('#' + form_id + ' .show-overlays').on('change', function() {
            price_chart.set('overlays', $('#' + form_id + ' .show-overlays').prop('checked'));
        });

        // Enable/Disable Annotations
        $('#' + form_id + ' .show-annotations').on('change', function() {
            price_chart.set('annotations', $('#' + form_id + ' .show-annotations').prop('checked'));
        });

        /*
        // Enable/Disable Legend
        $( '#' + form_id + ' .show-legend' ).on('change', function()
        {
            price_chart.set('show_legend', $( '#' + form_id + ' .show-legend' ).prop('checked') );
        });
        */

        // Enable/Disable TrendLines
        $('#' + form_id + ' .enable_trendlines').on('change', function() {
            if ($(this).is(':checked')) {
                price_chart.enterDrawMode();
                price_chart.external.mode = 'draw';
            } else {
                price_chart.enterDisplayMode();
                price_chart.external.mode = 'display';
            }
        });

        // Enable chart zooming
        if (_enable_zoom === true) {
            // Enable zoom buttons when mouse over chart
            $('.chart_canvas').mouseenter(function() {
                if (price_chart.external.mode != 'draw') {
                    price_chart.enterZoomMode();
                    price_chart.external.mode = 'zoom';
                }
            });

            // Disable zoom buttons when mouse out of chart
            $('.chart_canvas').mouseleave(function() {
                if (price_chart.external.mode == 'zoom') {
                    price_chart.enterDisplayMode();
                    price_chart.external.mode = 'display';
                }
            });
        }

        // Button to Reset Chart Zoom
        $('.reset').on('click', function() {
            $('.period').val('weekly').change();
            $('.enable_trendlines').removeAttr('checked');
            $('#moving_avg_1').removeAttr('checked').change();
            $('#moving_avg_2').removeAttr('checked').change();
            price_chart.reset();
        });

        // Enable/Disable Secondary Chart
        $('#' + form_id + ' .show-volume').on('change', function() {
            if ($('#' + form_id + ' .show-volume').prop('checked')) {
                $('.secondary_type').val('bar');
                $('.secondary_wrapper').val('volume');
                price_chart.set('secondary_chart_type', 'bar', false);
                price_chart.set('secondary_wrapper', 'volume', false);
            } else {
                $('..secondary_type').val('');
                $('.secondary_wrapper').val('');
                price_chart.set('secondary_chart_type', null, false);
                price_chart.set('secondary_wrapper', null, false);
            };

            price_chart.update();

            if (price_chart.error !== null)
                console.log('Could not build chart: ' + price_chart.error);
        });

        // Export Chart as PNG
        $('#' + form_id + ' .save-chart').on('click', function(e) {
            price_chart.save();
        });
    } else {
        if (window.charts.debug == true) {
            console.log('Unable to initialize chart');
            console.log('form_id: ' + form_id);
            console.log('chart_id: ' + chart_id);
        }
    }

    return price_chart;
}


// eof